import copy
import logging
from datetime import datetime
from os import path
from pathlib import Path
from xml.dom.minidom import Element, parse
from shutil import copyfile

import numpy as np
import pandas as pd
import rich_click as click
from PIL import Image as im
from rich import print
from rich.columns import Columns
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich_click import rich_config
from update_checker import UpdateChecker

from SimbioReader.constants import (CONTEXT_SETTINGS, FMODE, MSG, data_types,
                                    datamodel, progEpilog)
from SimbioReader.exceptions import SizeError
from SimbioReader.filters_tools import Filter
from SimbioReader.tools import (getElement, getValue, snake_case,gen_filename, lidUpdate, updateXML, lvidUpdate, pretty_print)
from SimbioReader.version import version

# from functools import wraps

click.rich_click.USE_RICH_MARKUP = True
click.rich_click.FOOTER_TEXT = progEpilog




__version__ = version.full()

    
def hdr_readr(file_name:Path):
    with open(file_name, FMODE.READ) as f:
        lines = f.readlines()
    img_data = {}
    info = ['samples', 'lines', 'bands', 'data type', 'Integration Time',
            'Compression Box', 'IBR', 'Start pixel row',
            'Start pixel col', 'Stop pixel row', 'Stop pixel col',
            'Binning W1']
    for line in lines:
        sect = line.split('=')
        if sect[0].strip() in info:

            if sect[0].strip() == 'Integration Time':
                img_data[snake_case(sect[0].strip())] = float(sect[-1].strip())
            elif sect[0].strip() == 'Compression Box':
                img_data[snake_case(sect[0].strip())] = sect[-1].strip()
            elif sect[0].strip() == 'data type':
                img_data[snake_case(sect[0].strip())] = int(sect[-1].strip())
            else:
                img_data[sect[0].strip()] = int(sect[-1].split('(')[0].strip())
    return img_data


class HK:
    """
    A class representing housekeeping data for a SIMBIO-SYS image.

    This class initializes various attributes from a pandas DataFrame containing
    housekeeping data and provides methods to display this information.

    Args:
        df (pd.DataFrame): A pandas DataFrame containing housekeeping data.

    Attributes:
        df (pd.DataFrame): The DataFrame containing housekeeping data.
    """
    
    def __init__(self,df:pd.DataFrame):
        """
        Initializes the HK object by extracting information from the DataFrame.

        Args:
            df (pd.DataFrame): A pandas DataFrame containing housekeeping data.
        """
        self.df=df
        for i in df.columns:
            if type(df[i].values[0]) is str:
                val = df[i].values[0].strip()
            else:
                val = df[i].values[0]
            setattr(self, i.strip().lower(), val)
            
    def Show(self) -> Panel:
        """
        Displays the housekeeping information in a formatted table.

        Returns:
            Panel: A Panel object containing the formatted housekeeping information.
        """
        sep=' = '
        dt=Table.grid()
        dt.add_column(style='yellow',justify='right')
        dt.add_column()
        dt.add_column(style='cyan',justify='left')
        for i in self.df.columns:
            dt.add_row(' '.join(i.split('_')).title(),sep,f"{self.df[i].values[0]}".strip())
        return Panel(dt,title='HouseKeeping',border_style='yellow',expand=False)
    
    def __str__(self)->str:
        """
        Returns a string representation of the HK object.

        Returns:
            str: A string representation of the HK object.
        """
        return f"HK object"
    
    def __repr__(self) -> str:
        """
        Returns a string representation of the HK object for debugging.

        Returns:
            str: A string representation of the HK object.
        """
        return self.__str__()
        


class Detector:
    """
    A class representing a detector in a SIMBIO-SYS image.

    This class extracts and initializes various attributes related to a detector
    in a SIMBIO-SYS image, such as the first line, first sample, and number of lines.

    Args:
        dat (Element): An XML Element containing the detector information.

    Attributes:
        first_line (int): The first line number of the detector.
        first_sample (int): The first sample number of the detector.
        lines (int): The number of lines in the detector.
    """
    def __init__(self,dat:Element) -> None:
        detector = getElement(dat, 'img:Detector')
        self.first_line = int(getValue(detector,'img:first_line'))
        self.first_sample = int(getValue(detector,'img:first_sample'))
        self.lines = int(getValue(detector,'img:lines'))
    
    def __str__(self) -> str:
        """
        Returns a string representation of the Detector object.

        Returns:
            str: A string representation of the Detector object.
        """
        return f"Detector object"
    
    def __repr__(self) -> str:
        """
        Returns a string representation of the Detector object for debugging.

        Returns:
            str: A string representation of the Detector object.
        """
        return self.__str__()
    
    def Show(self)-> Panel:
        """
        Displays the detector information in a formatted table.

        Returns:
            Panel: A Panel object containing the formatted detector information.
        """
        sep=' = '
        dt=Table.grid()
        dt.add_column(style='yellow',justify='right')
        dt.add_column()
        dt.add_column(style='cyan',justify='left')
        dt.add_row('First Line',sep,str(self.first_line))
        dt.add_row('First Sample',sep,str(self.first_sample))
        dt.add_row('Lines',sep,str(self.lines))
        return Panel(dt,title='Detector',border_style='yellow',expand=False)
    
class SubFrame:
    """
    A class representing a subframe of a SIMBIO-SYS image.

    This class extracts and initializes various attributes related to a subframe
    of a SIMBIO-SYS image, such as the first line, first sample, number of lines,
    number of samples, and field of view.

    Args:
        dat (Element): An XML Element containing the subframe information.

    Attributes:
        first_line (int): The first line number of the subframe.
        first_sample (int): The first sample number of the subframe.
        lines (int): The number of lines in the subframe.
        samples (int): The number of samples in the subframe.
        line_fov (float): The line field of view of the subframe.
        sample_fov (float): The sample field of view of the subframe.
    """
    def __init__(self, dat:Element) -> None:
        subFrame = getElement(dat, 'img:Subframe')
        self.first_line = int(getValue(subFrame,'img:first_line'))
        self.first_sample = int(getValue(subFrame,'img:first_sample'))
        self.lines = int(getValue(subFrame,'img:lines'))
        self.samples = int(getValue(subFrame,'img:samples'))
        self.line_fov = float(getValue(subFrame,'img:line_fov'))
        self.sample_fov = float(getValue(subFrame,'img:sample_fov'))
    
    def __str__(self):
        """
        Returns a string representation of the Subframe object.

        Returns:
            str: A string in the format 'SubFrame object'.
        """
        return f"SubFrame object"
    
    def __repr__(self) -> str:
        """
        Returns a string representation for debugging purposes.

        Returns:
            str: A string in the format 'SubFrame object'.
        """
        return self.__str__()
    

class DataStructure:
    """
    A class representing the data structure of a SIMBIO-SYS file.

    This class extracts and initializes various attributes related to the data 
    structure of a SIMBIO-SYS file, such as creation time, file size, and axes 
    configuration, based on an XML Element.

    Args:
        dat (Element): An XML Element containing the data structure information.
        channel (str): The channel identifier (e.g., 'HRIC', 'STC', 'VIHI').

    Attributes:
        creation_time (datetime): The creation time of the file.
        file_size (int): The size of the file.
        md5 (str): The MD5 checksum of the file.
        axes (int): The number of axes in the data structure.
        band (int): The band information (default is 1).
        data_type (str): The type of data (e.g., 'UnsignedLSB2', 'IEEE754LSBSingle').

    """
    
    def __init__(self, dat:Element, channel:str):
        fao=getElement(dat,'File_Area_Observational')
        fl=getElement(fao,'File')
        self.creation_time = datetime.strptime(getValue(fl,'creation_date_time'),"%Y-%m-%d")
        self.file_size = int(getValue(fl,'file_size'))
        self.md5 = getValue(fl,'md5_checksum')
        self.axes = int(getValue(fao,'axes'))
        self.band= None
        if self.axes == 3 and channel != 'VIHI':
            raise ValueError("The number of axes is wrong for the channel")
        for i in range(self.axes):
            axis = getElement(fao,'Axis_Array',i)
            setattr(self,getValue(axis,'axis_name').lower(),int(getValue(axis,'elements')))
        if self.axes == 3:
            dat=getElement(fao,'Array_3D_Spectrum')
        elif self.axes == 2:
            dat=getElement(fao,'Array_2D_Image')
        self.data_type = getValue(dat,'data_type')
        if self.band is None:
            self.band = 1
        
        
    def __str__(self)->str:
        """
        Returns a string representation of the Datastructure.

        Returns:
            str: A string in the format 'DataStructure Object'.
        """
        return f"DataStructure object"
    
    def __repr__(self) -> str:
        """
        Returns a string representation for debugging purposes.

        Returns:
            str: A string in the format 'DataStructure Object'.
        """
        return self.__str__()
    
    def Show(self):
        """
        Displays the data structure information in a formatted table.

        Returns:
            Panel: A rich Panel object containing the formatted table of data structure information.
        """
        sep=' = '
        dt=Table.grid()
        dt.add_column(style='yellow',justify='right')
        dt.add_column()
        dt.add_column(style='cyan',justify='left')
        for item in self.__dict__:
            dt.add_row(item,sep,str(self.__dict__[item]))
        # dt.add_row('Creation Time',sep,datetime.strftime(self.creation_time,"%Y-%m-%d"))
        # dt.add_row('File Size',sep,str(self.file_size))
        # dt.add_row('MD5 Checksum',sep,self.md5)
        # dt.add_row('Axes',sep,str(self.axes))
        return Panel(dt,title='Data Structure',border_style='yellow',expand=False)
        

class SimbioReader:
    """
        SimbioReader class for reading and processing SIMBIO-SYS data files.

        This class provides methods to read SIMBIO-SYS data files (STC, HRIC, VIHI channels),
        extract metadata, and access image data. It also offers functionalities
        for displaying information and saving image previews.

        Args:
            fileName (Path): Path to the Simbio data file.
            log (logging.Logger, optional): Logging object for recording messages. Defaults to None.
            debug (bool, optional): Flag to enable debug messages. Defaults to False.
            verbose (bool, optional): Flag to enable verbose messages. Defaults to False.
            console (Console, optional): Console object for printing messages. Defaults to None.
            updateCheck (bool, optional): Flag to check for updates on class initialization. Defaults to True.

        Attributes:
            fileName (Path): Path to the Simbio data file.
            log (logging.Logger, optional): Logging object for recording messages. Defaults to None.
            debug (bool, optional): Flag to enable debug messages. Defaults to False.
            verbose (bool, optional): Flag to enable verbose messages. Defaults to False.
            console (Console, optional): Console object for printing messages. Defaults to None.
            updateCheck (bool, optional): Flag to check for updates on class initialization. Defaults to True.
            channel (str): Channel type of the Simbio data (STC, HRIC, VIHI).
            pdsLabel (Path, optional): Path to the PDS4 label file.
            hk (HK, optional): Instance of the HK class containing housekeeping data (if available).
            title (str): Title of the Simbio data acquisition.
            dataModelVersion (str): Data model version of the Simbio data.
            startTime (datetime): Start time of the Simbio data acquisition.
            stopTime (datetime): Stop time of the Simbio data acquisition.
            level (str): Processing level of the Simbio data.
            tarName (str): Name of the target observed in the Simbio data.
            tarType (str): Type of the target observed in the Simbio data.
            startScet (str): Start spacecraft clock count of the Simbio data acquisition.
            stopScet (str): Stop spacecraft clock count of the Simbio data acquisition.
            phaseName (str): Mission phase name associated with the Simbio data acquisition.
            exposure (float): Exposure duration of the Simbio data acquisition (in seconds).
            firstLine (int): First line number of the image data.
            firstSample (int): First sample number of the image data.
            lines (int): Number of lines in the image data.
            samples (int): Number of samples in the image data.
            lineFov (float): Line field of view of the image data (in units).
            sampleFov (float): Sample field of view of the image data (in units).
            filter (Filter, optional): Instance of the Filter class containing filter information (if applicable).
            data_structure (DataStructure): Instance of the DataStructure class containing information about the data structure.
            img (np.ndarray): NumPy array containing the image data.
        """
    

    def __init__(self, fileName: Path, log: logging = None, 
                 debug: bool = False, verbose: bool = False,
                 console:Console=None, updateCheck:bool = True):

        if console is None:
            # from rich.console import Console
            self.console=Console()
        else:
            self.console=console
        if updateCheck:
            checker = UpdateChecker()
            result = checker.check('SimbioReader', version.short())
            if result:
                # TODO: Check this part after the delivery on pypi.org
                self.console.print(result)
                
        self._dateformat = "%Y-%m-%dT%H:%M:%S.%fZ"
        if not isinstance(fileName, Path):
            fileName = Path(fileName).resolve()
        if 'stc' in fileName.stem:
            self.channel = "STC"
        elif 'hric' in fileName.stem:
            self.channel = "HRIC"
        elif 'vihi' in fileName.stem:
            self.channel = "VIHI"
        else:
            raise ValueError('The file is not a Simbio file')
        
        self.fileName = fileName.absolute()
        self.log = log
        self.debug = debug
        self.verbose = verbose
        if self.fileName.suffix in ['.dat','.qub']:
            if self.fileName.with_suffix('.lblx').exists():
                self.pdsLabel=self.fileName.with_suffix('.lblx')
            else:
                raise ValueError("The PDS4 label is not present. it's impossible read the data file")
        if self.fileName.suffix == '.qub' and self.channel != 'VIHI':
            raise ValueError("The file is not a Simbio file or the filename is not well formatted")
        if self.fileName.suffix == '.lblx':
            if self.fileName.with_suffix('.dat').exists():
                self.pdsLabel=copy.copy(self.fileName)
                self.fileName=self.fileName.with_suffix('.dat')
            elif self.fileName.with_suffix('.qub').exists():
                self.pdsLabel=copy.copy(self.fileName)
                self.fileName=self.fileName.with_suffix('.qub')
            else:
                raise ValueError("The PDS4 data file is not present. it's impossible read the data file")
        
        self.read()

    def read(self):
        """Reads and processes the SIMBIO-SYS file, extracting metadata and image data.

        Raises:
            ValueError: If the file is not a valid SIMBIO-SYS file, PDS4 label is missing,
                        or data file is not found.
            SizeError: If the computed file size based on image data and data type
                        does not match the actual file size.
        """
        message = f"Reading {self.fileName}"
        if self.log:
            self.log.info(message)
        if self.verbose:
            self.console.print(f"{MSG.INFO}{message}")
                
        # Check the existence of the ausiliary files        
        if not self.fileName.with_suffix('.csv').exists():
            message = f"The HK file do not exists"
            if self.log:
                self.log.warning(message)
            if self.verbose:
                self.console.print(f"{MSG.WARNING}{message}")
        else:
            if self.verbose:
                self.console.print(f"{MSG.INFO}Read the HK from the csv")
            df = pd.read_csv(self.fileName.with_suffix('.csv'), sep=',', header=0)
            self.hk = HK(df)
            
        doc= parse(self.pdsLabel.as_posix())
        idArea = getElement(doc, 'Identification_Area')
        self.title = getValue(idArea, 'title')
        self.dataModelVersion = getValue(idArea, 'information_model_version')
        self.version =getValue(idArea, 'version_id')
        self.lid = getValue(idArea, 'logical_identifier')
        # TODO: Check on the datamodel version
        
        obsArea = getElement(doc, 'Observation_Area')
        timeCoords = getElement(obsArea, 'Time_Coordinates')
        self.startTime = datetime.strptime(getValue(timeCoords, 'start_date_time'),self._dateformat)
        self.stopTime = datetime.strptime(getValue(timeCoords, 'stop_date_time'),self._dateformat)
        
        primResSum = getElement(obsArea, 'Primary_Result_Summary')
        self.level = getValue(primResSum, 'processing_level')
        
        targetInfo = getElement(obsArea, 'Target_Identification')
        self.tarName = getValue(targetInfo, 'name')
        self.tarType = getValue(targetInfo, 'type')
        
        missionArea = getElement(doc, 'Mission_Area')
        missionInfo = getElement(missionArea, 'psa:Mission_Information')
        self.startScet = getValue(missionInfo, 'psa:spacecraft_clock_start_count')
        self.stopScet = getValue(missionInfo, 'psa:spacecraft_clock_stop_count')
        mission_phase=getElement(doc,"psa:Mission_Phase")
        self.phaseName = getValue(
            mission_phase, 'psa:name')
        
        discArea = getElement(doc, 'Discipline_Area')
        
        expInfo = getElement(discArea, 'img:Exposure')
        self.exposure = float(getValue(expInfo, 'img:exposure_duration'))/1000.
        
        subFrame = getElement(discArea, 'img:Subframe')
        self.firstLine = int(getValue(subFrame, 'img:first_line'))
        self.firstSample = int(getValue(subFrame, 'img:first_sample'))
        self.lines = int(getValue(subFrame, 'img:lines'))
        self.samples = int(getValue(subFrame, 'img:samples'))
        self.lineFov = float(getValue(subFrame, 'img:line_fov'))
        self.sampleFov = float(getValue(subFrame, 'img:sample_fov'))
        
        if self.channel != 'VIHI':
            flt = getElement(discArea, 'img:Optical_Filter')
            self.filter=Filter(channel=self.channel,name=getValue(flt,'img:filter_name'))
        
        
        self.data_stucture = DataStructure(doc, self.channel)
        self.samples=self.data_stucture.sample
        self.lines=self.data_stucture.line
        self.bands=self.data_stucture.band
        
        self.detector = Detector(doc)
        
        if self.data_stucture.data_type == "UnsignedLSB2":
            dtype = np.int16
        elif self.data_stucture.data_type == "IEEE754LSBSingle":
            dtype = np.float32

        if self.verbose:
            self.console.print(f"{MSG.INFO}Loading: {self.fileName}")
            if self.data_stucture.axes == 3:
                self.console.print(f"{MSG.INFO}Image size: {self.samples}x{self.lines}x{self.bands}")
                imgSize = self.samples*self.lines*self.bands*data_types[self.data_stucture.data_type]['bits']
            else:
                self.console.print(f"{MSG.INFO}Image size: {self.samples}x{self.lines}")
                imgSize = self.samples*self.lines*data_types[self.data_stucture.data_type]['bits']
            
            self.console.print(f"{MSG.INFO}File size: {self.fileName.stat().st_size*8}")
            self.console.print(
                f"{MSG.INFO}Computed File Size: {imgSize}")
            if self.fileName.stat().st_size*8 != imgSize:
                raise SizeError(self.fileName.stat().st_size *
                                8,imgSize)
        #print(img_data['samples'],img_data['lines'],img_data['bands'])
        if self.data_stucture.axes == 3:
            self.img = np.fromfile(self.fileName, dtype=dtype,
                               count=self.samples*self.lines*self.bands)
        else:
            self.img = np.fromfile(self.fileName, dtype=dtype,
                                   count=self.samples*self.lines)
        if self.data_stucture.axes == 3: 
            if self.lines==1:
                self.img.shape = ( self.samples,self.bands)
            else:
                self.img.shape = (self.lines,self.samples, self.bands )
        else:
            self.img.shape = (self.samples, self.lines)
        if self.verbose:
            self.console.print(f"{MSG.INFO}Dimension of the old image array: {self.img.ndim}")
            # print(f"Size of the old image array: {self.img.size}")

    @property
    def lvid(self)->str:
        """Returns the LIDVID of the SIMBIO-SYS file.

        Returns:
            str: The LIDVID of the SIMBIO-SYS file.
        """
    
        return f"{self.lid}::{self.version}"

    def Show(self, hk:bool=False, detector:bool=False, data_structure:bool=False, all_info:bool=False)->Columns:
        """Displays information about the loaded SIMBIO-SYS file.

        Args:
            hk (bool, optional): If True, displays the housekeeping data. Defaults to False.
            detector (bool, optional): If True, displays the detector information. Defaults to False.
            data_structure (bool, optional): If True, displays the data structure information. Defaults to False.
            all_info (bool, optional): If True, displays all available information (equivalent to setting
                                        hk, detector, and data_structure to True). Defaults to False.

        Returns:
            Columns: A rich console object containing the formatted information tables.
        """
        sep=' = '
        sep2 = ' : '
        if all_info:
            hk=True
            detector=True
            data_structure=True
        
        info=Table.grid()
        info.add_column(style='yellow', justify='right')
        info.add_column()
        info.add_column(style='cyan', justify='left')
        info.add_row('Title',sep2,self.title)
        info.add_row('DataModel Version', sep2, self.dataModelVersion)
        info.add_row('Start Acquisition', sep2, datetime.strftime(self.startTime,self._dateformat))
        info.add_row('Stop Acquisition', sep2, datetime.strftime(self.stopTime, self._dateformat))
        info.add_row('Start Acquisition SCET',sep2, self.startScet)
        info.add_row("Stop Acquisition SCET", sep2, self.stopScet)
        info.add_row("Phase Name", sep2, self.phaseName)
        
        infoT=Table.grid()
        infoT.add_column()
        infoT.add_row(Panel(info,title='General Info',border_style='yellow',expand=False))
        
        g=Table.grid()
        g.add_column()
        g.add_row(Panel(Text(self.channel, justify='center', style='magenta'), border_style='yellow', title="Channel"))
        g.add_row(Panel(Text(self.level.upper(), justify='center',
                  style='magenta'), border_style='yellow', title="Level"))
        
        tbs=Table.grid()
        tbs.add_column(style='yellow',justify='right')
        tbs.add_column()
        tbs.add_column(style='cyan',justify='left')
        tbs.add_row('Samples',sep,str(self.samples))
        tbs.add_row('Lines',sep,str(self.lines))
        if self.channel == 'VIHI':
            tbs.add_row('Bands',sep,str(self.bands))
        tbs.add_section()
        tbs.add_row("First Line", sep, str(self.firstLine))
        tbs.add_row("First Sample", sep, str(self.firstSample))
        tbs.add_row('Line FOV',sep, str(self.lineFov))
        tbs.add_row('Sample FOV', sep, str(self.sampleFov))
        
        # g.add_row(Panel(tbs, title='Image Info',
        #           border_style='yellow', expand=True))
        
        t=Table.grid()
        t.add_column(style='yellow', justify='right')
        t.add_column()
        t.add_column(style='cyan', justify='left')
        t.add_row("Name",sep2,self.tarName)
        t.add_row("Type", sep2, self.tarType)
        
        g.add_row(Panel(t, title='Target Info',
                  border_style='yellow'))
        
        g.add_row(Panel(tbs, title='Image Info',
                  border_style='yellow', expand=True))
        
 

        # phk=Panel(hk,title='HouseKeeping',border_style='yellow',expand=False)
        show_list = [g, infoT]
        if hk :
            show_list.append(self.hk.Show())
        
        if detector:
            g.add_row(self.detector.Show())
            
        if data_structure:
            infoT.add_row(self.data_stucture.Show())
            
        if self.channel != 'VIHI':
            
            infoT.add_row(self.filter.show())
   
        return Columns(show_list, title=self.fileName.stem)
        pass

    def __str__(self)-> str:
        """
        Returns a string representation of the object.

        Returns:
            str: A string in the format 'SimbioReader object <version> - from file <fileName>'.
        """
        return f"SimbioReader object {version} - from file {self.fileName.stem}"
    
    def __repr__(self) -> str:
        """
        Returns a string representation for debugging purposes.

        Returns:
            str: A string in the format 'SimbioReader object <version> - from file <fileName>'.
        """
        return self.__str__()
 

        
    def savePreview(self,img_type:str='png',quality:int=100, outFolder:Path=None, template:Path=None, description:str = "This is the first version.")->str|None:
        """Saves a preview image of the loaded data.

        Args:
            img_type: The desired image format. Supported formats are 'png', 'tif', and 'jpg'. Defaults to 'png'.
            quality: The quality of the saved image (applicable for 'png' and 'jpg' formats only). Ranges from 0 (worst) to 100 (best). Defaults to 100.
            out_folder: The output folder path where the preview image will be saved. Defaults to the same directory as the original Simbio file.
            template: name of the PDS4 template that will be generated. If None no template will be written. If the template is not none the img_type
                    is forced to png

        Raises:
            ValueError: If the provided image format is not supported.
            TypeError: If the `out_folder` argument is not a `Path` object.
        """
        if template:
            img_type = 'png'
        if outFolder is None:
            dest = self.fileName.parent
        else:
            if type(outFolder) is not Path:
                outFolder = Path(outFolder)
            dest=outFolder
        if img_type in ['png','tif']:
            data = im.fromarray(self.img)
            # print(data.getpixel((50, 50)))
            new_filename=gen_filename(self.fileName)
            # data_filename=self.fileName.stem
            # new_filename=copy.copy(data_filename)
            # if 'raw' in data_filename:
            #     new_filename.replace('raw','browse_raw')
            image_file= f"{dest}/{new_filename}.{img_type}"
            if 'cal' in self.fileName.stem:
                data.convert('RGB').save(image_file, quality=quality)
            else:
                data.save(image_file, quality=quality)
            if template:
                if not isinstance(template,Path):
                    template = Path(template)
                if not template.exists():
                    raise FileNotFoundError(f"The template {template.name} was not found")
                new_label=f"{dest}/{new_filename}.lblx"
                template.rename(new_label)
                from xml.dom.minidom import parse, parseString, Element
                import hashlib
                tree = parse(new_label)
                if 'cal' in self.fileName.stem:
                    calib=True
                else:
                    calib=False
                new_lid=lidUpdate(tree, new_label,calib=calib)
                creatTime=datetime.now()
                updateXML(tree, "modification_date", creatTime.strftime("%Y-%m-%d"), idx=0)
                file_version=new_filename.split('__')[1].split('.')[0]
                file_version=file_version.replace('_','.')
                updateXML(tree,"version_id",file_version, idx=0)
                updateXML(tree,"version_id",file_version, idx=1)
                updateXML(tree,"description",description, idx=0)
                lvidUpdate(tree,self.fileName, file_version)
                image_file= Path(image_file)
                updateXML(tree, 'file_name', image_file.name, idx=0)
                updateXML(tree, 'creation_date_time',
                    creatTime.strftime("%Y-%m-%d"), idx=0)
                updateXML(tree, 'file_size', image_file.stat().st_size, idx=0)
                updateXML(tree, 'md5_checksum', hashlib.md5(
                    open(image_file, 'rb').read()).hexdigest(), idx=0)
                dom2 = parseString(pretty_print(tree))
                with open(new_label, "w") as xmlFile:
                    dom2.writexml(xmlFile, encoding="utf-8")
                return f"{new_lid}::{file_version}"
        elif img_type == 'jpg':
            data = im.fromarray(self.img,mode='L')
            # print(data.getpixel((50,50)))
            data.save(f"{dest}/{self.fileName.stem}.{img_type}",
                      quality=quality)
        # print(self.img[0,0])
        pass
    
    def image(self)->im:
        """Returns a PIL Image object representing the loaded image data.

        This method returns a Pillow (PIL Fork) Image object containing the image data
        loaded from the Simbio file. The image data is assumed to be a single-band
        or multi-band array, depending on the channel type.

        Returns:
            A PIL Image object representing the loaded image data.

        Raises:
            ValueError: If the loaded image data cannot be converted to a PIL Image
            object due to unsupported data type or shape.
        """
        data = im.fromarray(self.img)
        return data


def sh_version(ctx, self, value):
    """Display version and exit"""
    console = Console()
    console.print(f"SimbioReader version [blue bold]{version.short()}[/]")
    console.print(f"DataModel version [blue bold]{datamodel}[/]")
    ctx.exit()


@click.command(context_settings=CONTEXT_SETTINGS)
@rich_config(help_config={'header_text': f"SIMBIO-SYS Data Reader, version [blue]{version.short()}[/blue]"})
@click.argument('file', type=click.Path(exists=True), required=True)
@click.option('--hk', is_flag=True, help='Show the HouseKeeping', default=False)
@click.option('--detector', is_flag=True, help='Show the Detector', default=False)
@click.option('--data-structure', is_flag=True, help='Show the Data Structure', default=False)
@click.option('--all', 'all_info', is_flag=True, help='Show all the information', default=False)
@click.option('--filter', 'filter_flag', is_flag=True, help='Show the filters for the given channel', default=False)
@click.option('-d', '--debug', is_flag=True, help='Debug mode', default=False)
@click.option('-v', '--verbose', is_flag=True, help='Verbose mode', default=False)
@click.option('--version', is_flag=True, help='Show the version and exit', default=False,callback=sh_version)
def cli(file: str=None, hk: bool = False, detector: bool=False, data_structure: bool = False, all_info: bool = False, 
        filter_flag : bool =False, debug: bool = False, verbose: bool = False, version:bool = False):
    console = Console()
    dat = SimbioReader(file, console=console, debug=debug, verbose=verbose)
    console.print(dat.Show(hk=hk, detector=detector,
                  data_structure=data_structure, all_info=all_info))
    







# def common_options_sub(func):
#     """Common options for the commands"""
#     @wraps(func)
#     @click.option('-a', '--all', is_flag=True, help='Show all the phases', default=False)
#     @click.option('-d', '--date', type=str, help='Show the phase for the given date', default=None)
#     @click.option('-n', '--name', type=str, help='Show the phase for the given name', default=None)
#     def wrapper(*args, **kwargs):
#         return func(*args, **kwargs)
#     return wrapper

# @cli.command()
# @common_options_sub
# def phases(all:bool,date:str, name:str):
#     """Display required phase(s)"""
#     from SimbioReader.phases_tools import get_phase, Phase
#     console = Console()
#     if not date and not name:
#         all=True
#     if all:
#         console.print(Phase.show())
#     elif date:
#         console.print(get_phase(dt=date).show())
#     elif name:
#         console.print(get_phase(name=name).show())
        

# @cli.command()
# @common_options_sub
# def subphases(all: bool, date: str, name: str):
#     """Display required subphase(s)"""
#     from SimbioReader.phases_tools import get_subphase, SubPhase
#     console = Console()
#     if not date and not name:
#         all = True
#     if all:
#         console.print(SubPhase.show())
#     elif date:
#         console.print(get_subphase(dt=date).show())
#     elif name:
#         console.print(get_subphase(name=name).show())
        

# @cli.command()
# @click.option('-a', '--all', is_flag=True, help='Show all the phases', default=False)
# @click.option('-d', '--date', type=str, help='Show the phase for the given date', default=None)
# @click.option('-n', '--name', type=str, help='Show the phase for the given name', default=None)
# @click.option('-p', '--phase', type=str, help='Show the tests for the given phase', default=None)
# @click.option('-s', '--subphase', type=str, help='Show the tests for the given subphase', default=None)
# def test(all: bool=None, date: str=None, name: str=None,phase:str=None,subphase:str=None):
#     """Display required test(s)"""
#     from SimbioReader.phases_tools import Test
#     console = Console()
#     if not date and not name and not phase and not subphase:
#         all = True
#     if all:
#         console.print(Test.show_all())
#     elif date:
#         console.print(Test.show_all(date=date))
#     elif name:
#         # console.print(Test.show(key=name))
#         console.print(Test(name,subphase=subphase).show())
#     elif phase:
#         console.print(Test.show_all(phase=phase))
#     elif subphase:
#         console.print(Test.show_all(subphase=subphase))
        
# @cli.command("filters")
# @click.argument('channel', required=True)
# def filters_act(channel:str):
#     """Display the filters for the given channel"""
#     from SimbioReader.filters_tools import show_filters
#     console = Console()
#     console.print(show_filters(channel))
    
    

if __name__ == "__main__":
    cli()


#! /usr/bin/env python3
from datetime import datetime

from dateutil.parser import parse
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from SimbioReader.phases import phases
from SimbioReader.subphases import subphases
from SimbioReader.tests import tests

dateFormat = "%Y-%m-%d %H:%M:%S"
console = Console()


class Phase:
    """
    Represents a phase, defined by a name or a date.

    Attributes:
        name (str): The name of the phase.
        start (datetime): The start time of the phase.
        end (datetime): The end time of the phase.
        extended_name (str): The extended name of the phase (LPName).
    """
    def __init__(self, name:str=None, dt:datetime|str=None):
        """
        Initializes a Phase object.

        Args:
            name (str, optional): The name of the phase. Defaults to None.
            dt (datetime | str, optional): The date used to determine the phase. Defaults to None.

        Raises:
            ValueError: If neither name nor dt is provided, or if no phase is found.
        """
        if name:
            phase_data = phases.get(name)
            if not phase_data:
                raise ValueError(f"Phase '{name}' not found.")
        elif dt:
            if isinstance(dt, str):
                dt = parse(dt,ignoretz=True)
            phase_data = self._find_phase_by_date(dt)
            if not phase_data:
                raise ValueError("No phase found for the given date.")
        else:
            raise ValueError("You must provide a phase name or a date.")
        
        self.name = phase_data["name"]
        self.start = parse(phase_data["start"],ignoretz=True)
        self.end = parse(phase_data["end"],ignoretz=True)
        self.extended_name = phase_data["LPName"]

    def _find_phase_by_date(self, dt):
        """
        Finds a phase by a given date.

        Args:
            dt (datetime): The date to find the corresponding phase.

        Returns:
            dict | None: The phase data if found, otherwise None.
        """
        for phase_name, phase_data in phases.items():
            start = parse(phase_data["start"],ignoretz=True)
            end = parse(phase_data["end"],ignoretz=True)
            if start <= dt <= end:
                return phase_data
        return None

    def __str__(self)-> str:
        """
        Returns a string representation of the phase.

        Returns:
            str: The string representation of the phase.
        """
        return f"Phase {self.name}"

    def __repr__(self) -> str:
        """
        Returns the official string representation of the phase.

        Returns:
            str: The string representation of the phase.
        """
        return self.__str__()

    @staticmethod
    def show_all()->Table:
        """
        Displays a table of all phases with their names and start/end times.

        Returns:
            rich.table.Table: A table displaying all phases.
        """
        tb = Table(style="yellow")
        tb.add_column('Phase Name')
        tb.add_column('Start Time')
        tb.add_column('End Time')
        for phase_name, phase_data in phases.items():
            if phase_name == "None":
                continue
            tb.add_row(f"{phase_data['LPName']} ({phase_name})",
                       parse(phase_data['start']).strftime(dateFormat),
                       parse(phase_data['end']).strftime(dateFormat))
        return tb
    
    def show(self):
        """
        Displays a table showing the details of this phase.

        Returns:
            rich.table.Table: A table displaying the phase details.
        """
        tb = Table(style="yellow")
        tb.add_column('Phase Name')
        tb.add_column('Start Time')
        tb.add_column('End Time')
        tb.add_row(f"{self.extended_name} ({self.name})",
                   self.start.strftime(dateFormat),
                   self.end.strftime(dateFormat))
        return tb

# Funzione di utilità per ottenere una fase


def get_phase(name:str=None, dt:datetime| str=None)->Phase:
    """
    Utility function to obtain a Phase object by name or date.

    Args:
        name (str, optional): The name of the phase. Defaults to None.
        dt (datetime | str, optional): The date used to determine the phase. Defaults to None.

    Returns:
        Phase: The corresponding Phase object.
    """
    if isinstance(dt, str):
        dt = parse(dt, ignoretz=True)
    return Phase(name=name, dt=dt)



class SubPhase:
    """
    Represents a subphase, which is a subdivision of a phase.

    Attributes:
        name (str): The name of the subphase.
        start (datetime): The start time of the subphase.
        end (datetime): The end time of the subphase.
        phase (Phase): The parent phase of the subphase.
        extended_name (str): The extended name of the subphase (LPName).
    """
    
    def __init__(self, name:str=None , dt:datetime | str = None):
        """
        Initializes a SubPhase object.

        Args:
            name (str, optional): The name of the subphase. Defaults to None.
            dt (datetime | str, optional): The date used to determine the subphase. Defaults to None.

        Raises:
            ValueError: If neither name nor dt is provided, or if no subphase is found.
        """
        if name:
            subphase_data = subphases.get(name)
            if not subphase_data:
                raise ValueError(f"SubPhase {name} not found.")
        elif dt:
            if isinstance(dt, str):
                dt = parse(dt,ignoretz=True)
            subphase_data = self._find_subphase_by_date(dt)
            if not subphase_data:
                raise ValueError("No subphase found for the given date.")
        else:
            raise ValueError("You must provide a subphase name or a date.")
        self.name = subphase_data['name']
        self.start = parse(subphase_data['start'])
        self.end = parse(subphase_data['end'])
        self.phase = Phase(subphase_data['phase'])
        self.extended_name = subphase_data['LPName']

    def _find_subphase_by_date(self, dt):
        """
        Finds a subphase by a given date.

        Args:
            dt (datetime): The date to find the corresponding subphase.

        Returns:
            dict | None: The subphase data if found, otherwise None.
        """
        for subphase_name, subphase_data in subphases.items():
            start = parse(subphase_data["start"],ignoretz=True)
            end = parse(subphase_data["end"],ignoretz=True)
            if start <= dt <= end:
                return subphase_data
        return None
    
    def __str__(self) -> str:
        """
        Returns a string representation of the subphase.

        Returns:
            str: The string representation of the subphase.
        """
        return f"SubPhase {self.name}"

    def __repr__(self) -> str:
        """
        Returns the official string representation of the subphase.

        Returns:
            str: The string representation of the subphase.
        """
        return self.__str__()
    
    @staticmethod
    def show_all() -> Table:
        """
        Displays a table of all subphases with their names, phases, and start/end times.

        Returns:
            rich.table.Table: A table displaying all subphases.
        """
        tb = Table(style="yellow")
        tb.add_column('SubPhase Name')
        tb.add_column('Extended Name')
        tb.add_column('Phase')
        tb.add_column('Start Time')
        tb.add_column('End Time')
        for phase_name, phase_data in subphases.items():
            if phase_name == "None":
                continue
            tb.add_row(phase_name,
                       phase_data['LPName'],
                       phase_data['phase'],
                       parse(phase_data['start']).strftime(dateFormat),
                       parse(phase_data['end']).strftime(dateFormat))
        return tb
    
    def show(self) -> Table:
        """
        Displays a table showing the details of this subphase.

        Returns:
            rich.table.Table: A table displaying the subphase details.
        """
        tb = Table(style="yellow")
        tb.add_column('SubPhase Name')
        tb.add_column('Extended Name')
        tb.add_column('Phase')
        tb.add_column('Start Time')
        tb.add_column('End Time')
        tb.add_row(self.name,
                   self.extended_name,
                   self.phase.name,
                   self.start.strftime(dateFormat),
                   self.end.strftime(dateFormat))
        return tb

def get_subphase(name:str=None, dt:datetime|str =None)->SubPhase:
    """
    Utility function to obtain a SubPhase object by name or date.

    Args:
        name (str, optional): The name of the subphase. Defaults to None.
        dt (datetime | str, optional): The date used to determine the subphase. Defaults to None.

    Returns:
        SubPhase: The corresponding SubPhase object.
    """
    if isinstance(dt, str):
        dt = parse(dt, ignoretz=True)
    return SubPhase(name=name, dt=dt)

def get_subphases_by_phase(phase_name:str)->list[str]|str:
    """
    Returns a list of subphases for a given phase.

    Args:
        phase_name (str): The name of the phase.

    Returns:
        list[str] | str: A list of subphases if more than one, or a single subphase name if only one.
    
    Raises:
        ValueError: If no subphases are found for the given phase.
    """
    elem=[subphase_name for subphase_name, subphase_data in subphases.items() if subphase_data['phase'] == phase_name]
    if len(elem)==1:
        return elem[0]
    elif len(elem)==0:
        raise ValueError(f"No subphases found for the phase {phase_name}")
    else:
        return elem
    
def compare_str(str1:str, str2:str)->bool:
    """
    Compares two strings to see if the first string (or its components if split by spaces) is contained within the second string.

    Args:
        str1 (str): The first string to compare.
        str2 (str): The second string to compare against.

    Returns:
        bool: True if str1 is contained within str2, otherwise False.
    """
    if ' ' in str1:
        pieces = str1.split(' ')
        return all(piece.lower() in str2.lower() for piece in pieces)
    else:
        return str1.lower() in str2.lower()
     
class Test:
    """
    Represents a test, associated with a specific subphase.

    Attributes:
        name (str): The name of the test.
        start (datetime): The start time of the test.
        end (datetime): The end time of the test.
        subphase (str): The subphase associated with the test.
    """
    
    def __init__(self, name:str=None, dt:datetime|str=None, subphase:str=None):
        """
        Initializes a Test object.

        Args:
            name (str, optional): The name of the test. Defaults to None.
            dt (datetime | str, optional): The date used to determine the test. Defaults to None.
            subphase (str, optional): The subphase associated with the test. Required if name is provided. Defaults to None.

        Raises:
            ValueError: If neither name nor dt is provided, or if no test is found, or if multiple tests are found without a detailed name.
        """
        if name:
            test_data = tests.get(name)
            if not test_data:
                if not subphase:
                    raise ValueError(f"To search a Test by name a subphase name is required.")
                sub_list = get_test_by_subphase(subphase)
                if not sub_list:
                    raise ValueError(f"Test {name} not found.")
                else:
                    
                    itm=[tests[testname] for testname in sub_list if compare_str(name, tests[testname]['name'])]
                    if len(itm)>1:
                        raise ValueError(f"Multiple tests found for the subphase {subphase}. Please provide detaile the test name.")
                    elif len(itm)==1:
                        test_data = itm[0]
                    else:
                        raise ValueError(f"Test {name} not found.")

        elif dt:
            if isinstance(dt, str):
                dt = parse(dt,ignoretz=True)
            test_data = self._find_test_by_date(dt)
            if not test_data:
                raise ValueError("No test found for the given date.")
        else:
            raise ValueError("You must provide a test name or a date.")
        self.name = test_data['name']
        self.start = parse(test_data['start'])
        self.end = parse(test_data['end'])
        self.subphase = test_data['subphase']
        
    def _find_test_by_date(self, dt):
        """
        Finds a test by a given date.

        Args:
            dt (datetime): The date to find the corresponding test.

        Returns:
            dict | None: The test data if found, otherwise None.
        """
        for test_name, test_data in tests.items():
            start = parse(test_data["start"], ignoretz=True)
            end = parse(test_data["end"], ignoretz=True)
            # console.print(f"Start: {start}, End: {end}, Date: {dt}")
            if start <= dt <= end:
                return test_data
        return None
    
    def __str__(self) -> str:
        """
        Returns a string representation of the test.

        Returns:
            str: The string representation of the test.
        """
        return f"Test {self.name} on {self.subphase}"
    
    def __repr__(self) -> str:
        """
        Returns the official string representation of the test.

        Returns:
            str: The string representation of the test.
        """
        return self.__str__()
    
    def show(self) -> Table:
        """
        Displays a table showing the details of this test.

        Returns:
            rich.table.Table: A table displaying the test details.
        """
        tb = Table(style="yellow")
        tb.add_column('Test Name')
        tb.add_column('SubPhase')
        tb.add_column('Start Time')
        tb.add_column('End Time')
        tb.add_row(self.name,self.subphase,self.start.strftime(dateFormat),self.end.strftime(dateFormat))
        return tb
    
    @staticmethod
    def show_all(phase:str=None,subphase:str=None,key:str=None,date:datetime|str=None) -> Table:
        """
        Displays a table showing all tests that match the provided filters.

        Args:
            phase (str, optional): The phase to filter tests by. Defaults to None.
            subphase (str, optional): The subphase to filter tests by. Defaults to None.
            key (str, optional): A keyword to search within test names. Defaults to None.
            date (datetime | str, optional): A date to filter tests by. Defaults to None.

        Returns:
            rich.table.Table: A table displaying the filtered tests.
        """
        tb = Table(style="yellow")
        tb.add_column('Test Name')
        tb.add_column('SubPhase')
        tb.add_column('Start Time')
        tb.add_column('End Time')
        console.print(f"Phase: {phase}, SubPhase: {subphase}, Key: {key}, Date: {date}")
        if date:
            if isinstance(date, str):
                date = parse(date, ignoretz=True)
        for test_name, test_data in tests.items():
            if subphase and not test_data['subphase'].lower() == subphase.lower():
                continue

            if not date and key:
                if not compare_str(key, test_data['name']):
                    continue    

            if date :
                start = parse(test_data['start'], ignoretz=True)
                end = parse(test_data['end'], ignoretz=True)
                if not start <=date <=end:
                    continue
            if phase:
                if not test_data['subphase'].lower() in get_subphases_by_phase(phase):
                    continue
            tb.add_row(test_data['name'],
                       test_data['subphase'],
                       parse(test_data['start']).strftime(dateFormat),
                       parse(test_data['end']).strftime(dateFormat))
        return tb

def get_test_by_subphase(subphase:str)->list[str] | None:
    """
    Returns a list of tests for a given subphase.

    Args:
        subphase (str): The name of the subphase.

    Returns:
        list[str] | None: A list of test names, or None if no tests are found.
    """
    elem= [test_name for test_name, test_data in tests.items() if test_data['subphase'].lower() == subphase.lower()]
    if len(elem)==0:
        return None
    elif len(elem)==1:
        return elem[0]
    else:
        return elem
        

class Filter:
    """
    Represents a filter for a specific channel, either 'HRIC' or 'STC'.
    
    Attributes:
        channel (str): The channel associated with the filter.
        name (str): The name of the filter.
        Other attributes dynamically assigned based on filter data.
    """
    
    def __init__(self, channel: str, name: str):
        """
        Initializes a Filter object based on the specified channel and filter name.

        Args:
            channel (str): The channel for which the filter is associated. Must be 'HRIC' or 'STC'.
            name (str): The name of the filter to initialize.
        
        Raises:
            ValueError: If the channel is invalid, or if no/multiple filters are found with the given name.
        """
        if channel.lower() == "hric":
            from SimbioReader.filters import hricFilters
            flt = hricFilters
        elif channel.lower() == "stc":
            from SimbioReader.filters import stcFilters
            flt = stcFilters
        else:
            raise ValueError("Invalid channel.")
        itm = [elem for elem in flt.values() if elem['name'].lower() == name.lower()]
        if len(itm) == 0:
            raise ValueError(f"No filter found with the name {name}.")
        elif len(itm) > 1:
            raise ValueError(
                "Multiple filters found. Please provide a detailed filter name.")
        elif len(itm) == 1:
            itm = itm[0]

        for key, value in itm.items():
            setattr(self, key, value)
        if 'name' not in self.__dict__:
            raise ValueError("Filter not found.")

    def __str__(self):
        """
        Returns a string representation of the Filter object.

        Returns:
            str: A string in the format 'Filter <name>'.
        """
        return f"Filter {self.name}"

    def __repr__(self):
        """
        Returns a string representation for debugging purposes.

        Returns:
            str: A string in the format 'Filter <name>'.
        """
        return self.__str__()

    def show(self)->Panel:
        """
        Displays the filter's data in a formatted table.

        Returns:
            Panel: A rich Panel object containing a table with the filter's attributes and values.
        """
        tb = Table.grid()
        tb.add_column(style='yellow')
        sep = ' = '
        for key, value in self.__dict__.items():
            tb.add_row(key.title(), sep, value)
        return Panel(tb, title='Filter',
                     border_style='yellow', expand=False)


def show_filters(channel: str) -> Table:
    """
    Displays all filters for a specified channel in a table.

    Args:
        channel (str): The channel for which to show filters. Must be 'HRIC' or 'STC'.

    Raises:
        ValueError: If the channel is invalid.

    Returns:
        Table: A rich Table object with all filters for the specified channel.
    """
    if channel.lower() == "hric":
        from SimbioReader.filters import hricFilters
        flt = hricFilters
    elif channel.lower() == "stc":
        from SimbioReader.filters import stcFilters
        flt = stcFilters
    else:
        raise ValueError("Invalid channel.")
    tb = Table(style="yellow")
    elem = next(iter(flt.values()))
    mask = {"desc": "Description"}
    for item in elem.keys():
        tb.add_column(item.title() if item not in ['desc'] else mask[item])

    for name, item in flt.items():
        tb.add_row(*[item[key] for key in elem.keys()])

    return tb



class SizeError(Exception):
    def __init__(self, size, expected):
        self.size = size
        self.expected = expected
        super().__init__(self.size, self.expected)

    def __str__(self):
        return f"The file size is {self.size} but the expected size is {self.expected}"


from SimbioReader.simbioInfo import get_phase, get_subphase,get_subphases_by_phase,Test
from datetime import datetime

def test_pahse_get_by_name():
    test01 = get_phase(name='necp')
    assert test01.extended_name == 'Near-Earth Commissioning Phase'

def test_pase_date_type():
    test01 = get_phase(name='necp')
    assert isinstance(test01.start, datetime)
    
def test_pase_date_value():
    test01 = get_phase(name='necp')
    assert test01.start == datetime(2018, 12, 10, 0, 0, 0)


def test_pase_date_value():
    test01 = get_phase(name='necp')
    assert test01.start == datetime(2018, 12, 10, 0, 0, 0)
    
def test_pase_get_by_date_str():
    test01 = get_phase(dt='2018-12-11 00:00:00')
    assert test01.extended_name == 'Near-Earth Commissioning Phase'
    

def test_pase_get_by_date_datetime():
    test01 = get_phase(dt=datetime(2018,12,11,0,0,0))
    assert test01.extended_name == 'Near-Earth Commissioning Phase'
    
def test_subphase_name():
    test01 = get_subphase(name='necp')
    assert test01.extended_name == 'Near-Earth Commissioning Phase'
    

def test_subpase_date_type():
    test01 = get_subphase(name='ico1')
    assert isinstance(test01.start, datetime)
    

def test_subpase_date_value_fail():
    test01 = get_subphase(name='ico1')
    assert test01.start != datetime(2018, 12, 10, 0, 0, 0)
    
def test_get_subphases_by_phase():
    test01 = get_subphases_by_phase(phase_name='necp')
    assert test01 == 'necp'


def test_subpase_date_value():
    test01 = get_subphase(name='ico1')
    assert test01.start != datetime(2019, 6, 7, 10, 0, 0)


def test_subpase_get_by_date():
    test01 = get_subphase(dt='2019-06-07 10:30:00')
    assert test01.extended_name == 'Instrument Check-Out #1'
    
def test_get_test_by_name():
    test01 = Test('Hric performance',subphase='ico9')
    assert test01.name == 'HRIC Performance Test'



from rich.table import Table
from rich.panel import Panel


class Filter:
    def __init__(self,channel:str, name:str):
        """Filter class"""
        if channel.lower() == "hric":
            from SimbioReader.filters import hricFilters
            flt=hricFilters
        elif channel.lower() == "stc":
            from SimbioReader.filters import stcFilters
            flt=stcFilters
        else:
            raise ValueError("Invalid channel.")
        itm = [elem for elem in flt.values() if elem['name']==name]
        if len(itm)==0:
            raise ValueError("No filters found.")
        elif len(itm)>1:
            raise ValueError("Multiple filters found. Please provide a detailed filter name.")
        elif len(itm)==1:
            itm=itm[0]
  
        for key, value in itm.items():
            setattr(self, key, value)      
        if 'name' not in self.__dict__:
            raise ValueError("Filter not found.")
        
    def __str__(self):
        return f"Filter {self.name}"
    
    def __repr__(self):
        return self.__str__()
    
    def show(self):
        """Show the filter data in a table"""
        tb = Table.grid()
        tb.add_column(style='yellow')
        sep= ' = '
        for key, value in self.__dict__.items():
            tb.add_row(key.title(), sep, value)
        return Panel(tb,title='Filter',
                     border_style='yellow', expand=False)
        
    


def show_filters(channel:str)->Table:
    if channel.lower() == "hric":
        from SimbioReader.filters import hricFilters
        flt=hricFilters
    elif channel.lower() == "stc":
        from SimbioReader.filters import stcFilters
        flt=stcFilters
    else:
        raise ValueError("Invalid channel.")
    tb = Table(style="yellow")
    elem=next(iter(flt.values()))
    mask = {"desc":"Description"}
    for item in elem.keys():
        tb.add_column(item.title() if item not in ['desc'] else mask[item])

    for name, item in flt.items():
        tb.add_row(*[item[key] for key in elem.keys()])
        
    return tb


#! /usr/bin/env python3


stcFilters={
    "panl":{
        "name"         : "PAN-L",
        "desc"         : "Fraction of the CCD used by the PAN-L filter",
        "id"           : "L_P700",
        "bandwidth"    : "204.3",
        "centerFilter" : "699.5",
        "interframeDel": "0.8",
        "fistLine"     : "1227",
        "firstSample"  : "576",
        "line"         : "384",
        "sample"       : "896"
    },
    "panh":{
        "name"         : "PAN-H",
        "desc"         : "Fraction of the CCD used by the PAN-H filter",
        "id"           : "H_P700",
        "bandwidth"    : "204.3",
        "centerFilter" : "699.5",
        "interframeDel": "0.8",
        "fistLine"     : "1227",
        "firstSample"  : "576",
        "line"         : "384",
        "sample"       : "896"
    },
    "f920":{
          "name"         : "F920-L",
          "desc"         : "Fraction of the CCD used by the F920-L filter",
          "id"           : "L_F920",
          "bandwidth"    : "20",
          "centerFilter" : "920",
          "interframeDel": "0.8",
          "fistLine"     : "1953",
          "firstSample"  : "576",
          "line"         : "64",
          "sample"       : "896"
    },
    "f550":{
        "name"         : "F550-L",
        "desc"         : "Fraction of the CCD used by the F550-L filter",
        "id"           : "L_F550",
        "bandwidth"    : "20",
        "centerFilter" : "550",
        "interframeDel": "0.8",
        "fistLine"     : "1745",
        "firstSample"  : "576",
        "line"         : "64",
        "sample"       : "896"
    },
    "f420":{
        "name"         : "F420-H",
        "desc"         : "Fraction of the CCD used by the F420-H filter",
        "id"           : "H_F420",
        "bandwidth"    : "20",
        "centerFilter" : "420",
        "interframeDel": "0.8",
        "fistLine"     : "240",
        "firstSample"  : "576",
        "line"         : "64",
        "sample"       : "896"
    },
    "f750":{
        "name"         : "F750-H",
        "desc"         : "Fraction of the CCD used by the F750-H filter",
        "id"           : "H_F750",
        "bandwidth"    : "20",
        "centerFilter" : "750",
        "interframeDel": "0.8",
        "fistLine"     : "32",
        "firstSample"  : "576",
        "line"         : "64",
        "sample"       : "896"
    },
     "winx":{
        "name"         : "WIN-X",
        "desc"         : "Fraction of the CCD used for calibration",
        "id"           : "Win-X",
        "bandwidth"    : "0",
        "centerFilter" : "0",
        "interframeDel": "0.8",
        "fistLine"     : "100",
        "firstSample"  : "192",
        "line"         : "64",
        "sample"       : "128"
    },
}
hricFilters={
    "fpan":{
        "name"         : "PAN-L",
        "desc"         : "Fraction of the CCD used by the PAN-L filter",
        "id"           : "L_P700",
        "bandwidth"    : "204.3",
        "centerFilter" : "699.5",
        "interframeDel": "0.8",
        "fistLine"     : "1227",
        "firstSample"  : "576",
        "line"         : "384",
        "sample"       : "896"
    },
    "f550":{
        "name"         : "F550-L",
        "desc"         : "Fraction of the CCD used by the F550-L filter",
        "id"           : "L_F550",
        "bandwidth"    : "20",
        "centerFilter" : "550",
        "interframeDel": "0.8",
        "fistLine"     : "1745",
        "firstSample"  : "576",
        "line"         : "64",
        "sample"       : "896"
    },
    "f750":{
        "name"         : "F750-H",
        "desc"         : "Fraction of the CCD used by the F880-H filter",
        "id"           : "H_F420",
        "bandwidth"    : "20",
        "centerFilter" : "420",
        "interframeDel": "0.8",
        "fistLine"     : "240",
        "firstSample"  : "576",
        "line"         : "64",
        "sample"       : "896"
    },
    "f880":{
        "name"         : "F880-H",
        "desc"         : "Fraction of the CCD used by the F880-H filter",
        "id"           : "H_F750",
        "bandwidth"    : "20",
        "centerFilter" : "750",
        "interframeDel": "0.8",
        "fistLine"     : "32",
        "firstSample"  : "576",
        "line"         : "64",
        "sample"       : "896"
    },
    "cust0":{
        "name"         : "PAN-L - custom",
        "desc"         : "Fraction of the CCD used by the PAN-L filter",
        "id"           : "L_P700",
        "bandwidth"    : "204.3",
        "centerFilter" : "699.5",
        "interframeDel": "0.8",
        "fistLine"     : "1227",
        "firstSample"  : "576",
        "line"         : "384",
        "sample"       : "896"
    },
}



from SimbioReader.constants import VERSION


def shVersion():
    code = {"a": "alpha", "b": "beta", "rc": "ReleaseCandidate", "f": "Final"}
    nVer = f"{VERSION[0]}.{VERSION[1]}"
    if VERSION[2] != 0:
        nVer += f".{VERSION[2]}"
    if VERSION[3].lower() != "f":
        nVer += f".{code[VERSION[3].lower()]}"
        if VERSION[4] != 0:
            nVer += f".{VERSION[4]}"
    return nVer


code = {"d": "dev", "a": "alpha", "b": "beta",
        "rc": "ReleaseCandidate", "f": "Final"}


class Vers:
    def __init__(self, ver: tuple):
        self.major = ver[0]
        self.minor = ver[1]
        self.bug = ver[2]
        self.type = ver[3]
        self.build = ver[4]

    def full(self):
        if self.type.lower()=='f':
            nVer = f"{self.major}.{self.minor}.{self.bug}"
        else:
            nVer = f"{self.major}.{self.minor}.{self.bug}.{code[self.type]}.{self.build}"
        return nVer

    def short(self):
        return f"{self.major}.{self.minor}.{self.bug}"
    
    def __str__(self):
        return self.full()


version = Vers(VERSION)

# def get_version(version=None):
#     version = get_complete_version(version)
#     main = get_main_version(version)
#     sub = ''
#     if version[3] == 'alpha' and version[4] == 0:
#         git_changeset = get_git_changeset()
#         if git_changeset:
#             sub = f".dev{git_changeset}"
#     elif version[3] != 'final':
#         mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'rc'}
#         sub = f"{mapping[version[3]]}.{version[4]}"
#     return f"{main}.{sub}"


# def get_main_version(version=None):
#     version = get_complete_version(version)
#     parts = 2 if version[2] == 0 else 3
#     return '.'.join(str(x) for x in version[:parts])


# def get_complete_version(version=None):
#     if version is None:
#         version= VERSION
#     else:
#         assert len(version) == 5
#         assert version[3] in ('alpha', 'beta', 'rc', 'final')
#     return version


# @functools.lru_cache()
# def get_git_changeset():
#     """Return a numeric identifier of the latest git changeset.

#     The result is the UTC timestamp of the changeset in YYYYMMDDHHMMSS format.
#     This value isn't guaranteed to be unique, but collisions are very unlikely,
#     so it's sufficient for generating the development version numbers.
#     """
#     # Repository may not be found if __file__ is undefined, e.g. in a frozen
#     # module.
#     if '__file__' not in globals():
#         return None
#     repo_dir = path.dirname(path.dirname(path.abspath(__file__)))
#     git_log = subprocess.run(
#         'git log --pretty=format:%ct --quiet -1 HEAD',
#         stdout=subprocess.PIPE, stderr=subprocess.PIPE,
#         shell=True, cwd=repo_dir+'/..', universal_newlines=True,
#     )
#     timestamp = git_log.stdout
#     tz = timezone.utc
#     try:
#         timestamp = datetime.fromtimestamp(int(timestamp), tz=tz)
#     except ValueError:
#         return None
#     return timestamp.strftime('%Y%m%d%H%M%S')


# types = {
#     'd': 'dev',
#     'a': 'alpha',
#     'b': 'beta',
#     'f': 'final',
# }


# class Version:
#     """Version Class
#         respecting the semantic versioning 2.0.0
#     """

#     def __init__(self, version: tuple):
#         self.version = version

#     @property
#     def version(self):
#         if self._type is None:
#             adv = ""
#         else:
#             adv = f"-{self._type}"
#             if self._build is not None:
#                 adv += f".{self._build}"
#         return f"{self._major}.{self._minor}.{self._patch}{adv}"

#     @version.setter
#     def version(self, version):
#         tags = ('_major', '_minor', '_patch', '_type', '_build')
#         for tag in tags:
#             setattr(self, tag, 0)
#         # version=version.split(".")
#         for i in range(len(version[0:3])):
#             if type(version[i]) is str:
#                 if version[i].isdigit():
#                     setattr(self, tags[i], int(version[i]))
#                 else:
#                     raise ValueError(
#                         f"{tags[i][1:].title()} version must be a number")
#             elif type(version[i]) is float:
#                 setattr(self, tags[i], int(version[i]))
#             elif type(version[i]) is int:
#                 setattr(self, tags[i], version[i])
#         if len(version) > 3:
#             if len(version[3]) == 1:
#                 if not version[3] in types.keys():
#                     raise ValueError(f"Unknown type {version[3]}")
#                 else:
#                     self._type = types[version[3]]
#             else:
#                 if not version[3] in types.values():
#                     raise ValueError(f"Unknown type {version[3]}")
#                 self._type = version[3]
#         else:
#             self._type = None
#         if len(version) == 5:
#             self._build = version[4]
#         else:
#             self._build = None

#     def __str__(self) -> str:
#         if self._type == "final":
#             return f"{self._major}.{self._minor}.{self._patch}"
#         else:
#             return self.version

#     def full(self) -> str:
#         return self.version



class MSG:
    ERROR = "[red][ERROR][/red] "
    CRITICAL = "[red][CRITICAL][/red] "
    INFO = "[green][INFO][/green] "
    DEBUG = "[blue][DEBUG][/blue] "
    WARNING = "[yellow][WARNING][/yellow] "
    TODO = "[yellow][TODO][/yellow] "


class FMODE:
    READ = 'r'
    READ_BINARY = 'rb'
    WRITE = 'w'
    WRITE_BINARY = 'wb'
    APPEND = 'a'


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
progEpilog = "- For any information or suggestion please contact " \
    "[bold magenta]Romolo.Politi@inaf.it[/bold magenta]"

VERSION = (0,5,10,'d',1)

datamodel='1.22.0.0'

data_types = {
    "UnsignedLSB2": {'envi':2,'bits':16},
    "IEEE754LSBSingle":{'envi':4,'bits':32},
}



import xml.dom.minidom as md
from re import sub
from pathlib import Path
import copy

def getValue(nodeList: md.Element, label: str) -> str:
    """Get the value from a tag
    
    Args:
        nodelist: The xml block to evaluate
    
    Returns:
         The value of the tag

    """
    # for item in nodeList:
    #     print(item)
    elem = nodeList.getElementsByTagName(label)
    return elem[0].firstChild.data
    # return item


def getElement(doc, label, el=0) -> md.Element:
    """Get a Block of a dom
    
    Args:
        doc (xml.dom): The full Object
        
        label (str): The name of the tag to extract
            
    Returns:
        (xml.dom) The node tree extracted
        
    Todo:
        * implement OnBoard processing class
    """
    elem = doc.getElementsByTagName(label)
    return elem[el]


def camel_case(text)-> str:
    """
    Converts a given string to camelCase.

    The function transforms a string by removing underscores or hyphens,
    capitalizing the first letter of each word except the first one,
    and then joining them together to form a camelCase string.

    Args:
        text (str): The input string, typically in snake_case or kebab-case.

    Returns:
        str: The converted string in camelCase.

    Example:
        >>> camel_case('hello_world')
        'helloWorld'
        >>> camel_case('hello-world')
        'helloWorld'
    """
    text = sub(r"(_|-)+", " ", text).title().replace(" ", "")
    return ''.join([text[0].lower(), text[1:]])


def snake_case(text)->str:
    """
    Converts a given string to snake_case.

    The function transforms a camelCase or PascalCase string to snake_case,
    replacing capital letters with underscores followed by the lowercase equivalent.

    Args:
        text (str): The input string, typically in camelCase or PascalCase.

    Returns:
        str: The converted string in snake_case.

    Example:
        >>> snake_case('helloWorld')
        'hello_world'
        >>> snake_case('HelloWorld')
        'hello_world'
    """
    return '_'.join(
        sub('([A-Z][a-z]+)', r' \1',
            sub('([A-Z]+)', r' \1',
                text.replace('-', ' '))).split()).lower()


def gen_filename(old_filename:Path)->str:
    new_filename=copy.copy(old_filename.stem)
    # new_filename=new_filename.split('__')[0]
    if 'raw' in new_filename:
        new_filename=new_filename.replace('raw','browse_raw')
    else:
        new_filename=new_filename.replace('_cal_','_browse_cal_')
    return new_filename

def updateXML(xml: md.Element, label: str, value: str | int | float, idx: int = 0) -> None:
    a = xml.getElementsByTagName(label)[idx]
    a.firstChild.nodeValue = value
    


def getFromXml(xml:md.Element, label, idx=0):
    a = xml.getElementsByTagName(label)[idx]
    return a.firstChild.nodeValue

def lidGenerator(old_lid: str , file_name: Path, calib:bool=False)-> str:
    parts = old_lid.split(':')
    if calib:
        parts[-2] = 'browse_calibrated'
    parts[-1] = file_name.stem.split('__')[0]
    newLid = ':'.join(parts)
    return newLid

def lidUpdate(tree, fileName, calib: bool = False):
    # conf.log.debug("LID update", verbosity=3)
    oldLid = getFromXml(tree, "logical_identifier")
    # parts = oldLid.split(':')
    # if calib:
    #     parts[-2] = 'data_calibrated'
    # parts[-1] = fileName.stem.split('__')[0]
    # newLid = ':'.join(parts)
    if not isinstance(fileName, Path):
        fileName=Path(fileName)
    newLid =lidGenerator(old_lid=oldLid,file_name=fileName, calib=calib)
    updateXML(tree, "logical_identifier", newLid)
    return newLid
    

def new_lvid(old:str, file_name: Path, file_version:str):
    parts=old.split('::')
    new_main=parts[0].split(':')
    if 'cal' in file_name.stem:
        new_main[-2]='browse_calibrated'
    newLVID=f"{':'.join(new_main[0:-1])}:{file_name.stem.split('__')[0]}::{file_version}"
    return newLVID


def lvidUpdate(tree:str, file_name: Path, file_version:str):
    oldLVID = getFromXml(tree, "lidvid_reference")
    newLVID=new_lvid(oldLVID, file_name, file_version)
    updateXML(tree, "lidvid_reference", newLVID)
    
    pass

def pretty_print(dom):
    return '\n'.join([line for line in dom.toprettyxml(indent=' '*4).split('\n') if line.strip()])


from SimbioReader.tools import gen_filename, new_lvid
from pathlib import Path 

def test_filename():
    old_name=Path("sim_raw_sc_hric_cust0_internal_cruise_cruise_2021-04-24_001__0_1.dat")
    new_name=gen_filename(old_name)
    assert new_name == "sim_browse_raw_sc_hric_cust0_internal_cruise_cruise_2021-04-24_001__0_1"

def test_filename_cal():
    old_name=Path("sim_cal_sc_hric_cust0_internal_cruise_cruise_2021-04-24_001__0_1.dat")
    new_name=gen_filename(old_name)
    assert new_name == "sim_browse_cal_sc_hric_cust0_internal_cruise_cruise_2021-04-24_001__0_1"

def test_lvid():
    oldLVID="urn:esa:psa:bc_mpo_sim:data_raw:sim_raw_sc_hric_template::0.1"
    file_name=Path("sim_cal_sc_hric_cust0_internal_cruise_cruise_2021-04-24_001__0_1.dat")
    file_version='0.1'
    newLVID=new_lvid(oldLVID, file_name, file_version)
    assert newLVID == "urn:esa:psa:bc_mpo_sim:data_raw:sim_cal_sc_hric_cust0_internal_cruise_cruise_2021-04-24_001::0.1"


#!/usr/bin/env python3
import rich_click as click
from rich.console import Console
from functools import wraps
from SimbioReader.constants import FMODE, MSG, CONTEXT_SETTINGS, progEpilog, data_types, datamodel

click.rich_click.USE_RICH_MARKUP = True
click.rich_click.FOOTER_TEXT = progEpilog

version='0.1.0'
update='2024-09-01'

def show_version():
    """Display version and exit"""
    console = Console()
    console.print(f"SimbioInfo version [blue bold]{version}[/]")
    console.print(f"Last Update [blue bold]{update}[/]")


@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
@click.option('--version', 'ver', is_flag=True, help='Show the version', default=False)
@click.pass_context
def cli(ctx,ver: bool):
    """Simbio Info CLI"""
    if ver:
        show_version()
        ctx.exit()
    pass


def common_options_sub(func):
    """Common options for the commands"""
    @wraps(func)
    @click.option('-a', '--all', is_flag=True, help='Show all the phases', default=False)
    @click.option('-d', '--date', type=str, help='Show the phases for the given date', default=None)
    @click.option('-n', '--name', type=str, help='Show the phase with the given name', default=None)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper

@cli.command()
@common_options_sub
@click.pass_context
def phases(ctx,all: bool, date: str, name: str):
    """Display required phase(s)"""
    from SimbioReader.simbioInfo import get_phase, Phase
    from dateutil.parser import parse
    console = Console()
    if not date and not name:
        all = True

    if date:
        try:
            ndt = parse(date, ignoretz=True)
        except Exception as e:
            console.print(f"{MSG.ERROR}Cant convert to date the string: {str(e)}")
            all = True
    if all:
        console.print(Phase.show_all())
    elif date:
        try:
            console.print(get_phase(dt=date).show())
        except ValueError as e:
            ctx.fail(str(e))
    elif name:
        try:
            console.print(get_phase(name=name).show())
        except ValueError as e:
            ctx.fail(str(e))


@cli.command()
@common_options_sub
def subphases(all: bool, date: str, name: str):
    """Display required subphase(s)"""
    from SimbioReader.simbioInfo import get_subphase, SubPhase
    from dateutil.parser import parse
    console = Console()
    if not date and not name:
        all = True
    if date:
        try:
            ndt = parse(date, ignoretz=True)
        except Exception as e:
            console.print(
                f"{MSG.ERROR}Cant convert to date the string: {str(e)}")
            all = True
    if all:
        console.print(SubPhase.show_all())
    elif date:
        console.print(get_subphase(dt=date).show())
    elif name:
        console.print(get_subphase(name=name).show())


@cli.command()
@click.option('-a', '--all', is_flag=True, help='Show all the tests', default=False)
@click.option('-d', '--date', type=str, help='Show the tests for the given date', default=None)
@click.option('-n', '--name', type=str, help='Show the tests for the given name', default=None)
@click.option('-p', '--phase', type=str, help='Show the tests for the given phase', default=None)
@click.option('-s', '--subphase', type=str, help='Show the tests for the given subphase', default=None)
@click.pass_context
def tests(ctx,all: bool = None, date: str = None, name: str = None, phase: str = None, subphase: str = None):
    """Display required test(s)"""
    from SimbioReader.simbioInfo import Test
    console = Console()
    if not date and not name and not phase and not subphase:
        all = True
    if all:
        console.print(Test.show_all())
    elif date:
        try:
            console.print(Test(dt=date).show())
        except ValueError as e:
            ctx.fail(str(e))
    elif name:
        # console.print(Test.show(key=name))
        try:
            console.print(Test(name, subphase=subphase).show())
        except:
            if not subphase:
                console.print(f"{MSG.WARNING}Test {name.title()} not found. Please try to specific a subphase.")
                ctx.exit()
            console.print(f"{MSG.WARNING}Test {name.title()} for subphase {subphase.upper()} not found.")
            console.print(Test.show_all(key=name))
    elif phase:
        console.print(Test.show_all(phase=phase))
    elif subphase:
        console.print(Test.show_all(subphase=subphase))


@cli.command("filters")
@click.argument('channel', required=True)
@click.option('-n', '--name', type=str, help='Show the filter for the given name', default=None)
@click.pass_context
def filters_act(ctx, channel: str,name:str):
    """Display the filters for the given channel"""
    from SimbioReader.simbioInfo import show_filters
    console = Console()
    if name:
        from SimbioReader.simbioInfo import Filter
        try:
            fil=Filter(channel,name)
            console.print(fil.show())
        except ValueError as e:
            ctx.fail(str(e))
    else:
        console.print(show_filters(channel))


if __name__ == "__main__":
    cli()